@@ -21,7 +21,7 @@ from typing import Any, Dict, List, NoReturn, Optional, Tuple, Union
 
 import jsonschema
 from hypothesis.errors import InvalidArgument
-from hypothesis.internal.floats import next_down, next_up
+from hypothesis.internal.floats import next_down as ieee_next_down, next_up
 
 # Mypy does not (yet!) support recursive type definitions.
 # (and writing a few steps by hand is a DoS attack on the AST walker in Pytest)
@@ -51,6 +51,15 @@ SCHEMA_KEYS = tuple(
 SCHEMA_OBJECT_KEYS = ("properties", "patternProperties", "dependencies")
 
 
+def next_down(val: float) -> float:
+    """Compensate for JSONschema's lack of negative zero with an extra step."""
+    out = ieee_next_down(val)
+    if out == 0 and math.copysign(1, out) == -1:
+        out = ieee_next_down(out)
+    assert isinstance(out, float)
+    return out
+
+
 def make_validator(
     schema: Schema,
 ) -> Union[
@@ -143,10 +152,12 @@ def upper_bound_instances(schema: Schema) -> float:
         return len(schema["enum"])
     if get_type(schema) == ["integer"]:
         lower, upper = get_integer_bounds(schema)
-        mul = schema.get("multipleOf", 1)
-        assert isinstance(mul, (int, float))
-        if lower is not None and upper is not None and isinstance(mul, int):
-            return 1 + (upper - lower) % mul
+        if lower is not None and upper is not None:
+            mul = schema.get("multipleOf")
+            if isinstance(mul, int):
+                return 1 + (upper - lower) % mul
+            return 1 + (upper - lower)  # Non-integer mul can only reduce upper bound
+    # TODO: upper-bound array instances
     return math.inf
 
 
@@ -275,19 +286,19 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
             del schema["exclusiveMaximum"]
         lo, hi, exmin, exmax = get_number_bounds(schema)
         mul = schema.get("multipleOf")
-        if (
-            lo is not None
-            and hi is not None
-            and (
-                (mul and not has_divisibles(lo, hi, mul, exmin, exmax))
-                or (next_up(lo) if exmin else lo) > (next_down(hi) if exmax else hi)
-            )
-        ):
-            type_.remove("number")
-        elif isinstance(mul, int):
+        if isinstance(mul, int):
             # Numbers which are a multiple of an integer?  That's the integer type.
             type_.remove("number")
             type_ = [t for t in TYPE_STRINGS if t in type_ or t == "integer"]
+        elif lo is not None and hi is not None:
+            lobound = next_up(lo) if exmin else lo
+            hibound = next_down(hi) if exmax else hi
+            if (
+                mul and not has_divisibles(lo, hi, mul, exmin, exmax)
+            ) or lobound > hibound:
+                type_.remove("number")
+            elif type_ == ["number"] and lobound == hibound:
+                return {"const": lobound}
 
     if "integer" in type_:
         lo, hi = get_integer_bounds(schema)
@@ -297,17 +308,6 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         if hi is not None and isinstance(mul, int) and mul > 1 and (hi % mul):
             hi -= hi % mul
 
-        # TODO: this step leads to validation errors in the upstream jsonschema
-        #       library, due to floating-point imprecision.  Can we fix that?
-        #
-        # if "multipleOf" in schema:
-        #     mul = Fraction(schema["multipleOf"]).numerator
-        #     if lo is not None and mul > 1 and (lo % mul):
-        #         lo += mul - (lo % mul)
-        #     if hi is not None and mul > 1 and (hi % mul):
-        #         hi -= hi % mul
-        #     schema["multipleOf"] = mul
-
         if lo is not None:
             schema["minimum"] = lo
             schema.pop("exclusiveMinimum", None)
@@ -319,6 +319,11 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
             type_.remove("integer")
 
     if "array" in type_ and "contains" in schema:
+        if isinstance(schema.get("items"), dict):
+            contains_items = merged([schema["contains"], schema["items"]])
+            if contains_items is not None:
+                schema["contains"] = contains_items
+
         if schema["contains"] == FALSEY:
             type_.remove("array")
         else:
@@ -326,6 +331,8 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         if schema["contains"] == TRUTHY:
             schema.pop("contains")
             schema["minItems"] = max(schema.get("minItems", 1), 1)
+    # TODO: upper_bound_instances on the items, and use that to set maxItems
+    # constraint for unique arrays.  Not worth handling list-items though...
     if "array" in type_ and schema.get("minItems", 0) > schema.get(
         "maxItems", math.inf
     ):
@@ -333,7 +340,7 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
     if (
         "array" in type_
         and "minItems" in schema
-        and isinstance(schema.get("items", []), (bool, dict))
+        and isinstance(schema.get("items", []), dict)
     ):
         count = upper_bound_instances(schema["items"])
         if (count == 0 and schema["minItems"] > 0) or (
@@ -371,11 +378,12 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         schema.pop("items", None)
     if (
         "properties" in schema
+        and "maxProperties" in schema
         and not schema.get("patternProperties")
         and schema.get("additionalProperties") == FALSEY
     ):
         schema["maxProperties"] = min(
-            schema.get("maxProperties", math.inf), len(schema["properties"])
+            schema["maxProperties"], len(schema["properties"])
         )
     if "object" in type_ and schema.get("minProperties", 0) > schema.get(
         "maxProperties", math.inf
@@ -405,8 +413,11 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         schema["required"] = sorted(reqs)
         max_ = schema.get("maxProperties", float("inf"))
         assert isinstance(max_, (int, float))
+        properties = schema.get("properties", {})
         if len(schema["required"]) > max_:
             type_.remove("object")
+        elif any(properties.get(name, {}) == FALSEY for name in schema["required"]):
+            type_.remove("object")
         else:
             propnames = schema.get("propertyNames", {})
             validator = make_validator(propnames)
@@ -471,7 +482,10 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         if len(schema) == len(schema["anyOf"]) == 1:
             return schema["anyOf"][0]  # type: ignore
     if "allOf" in schema:
-        schema["allOf"] = sorted(schema["allOf"], key=encode_canonical_json)
+        schema["allOf"] = [
+            json.loads(enc)
+            for enc in sorted(set(map(encode_canonical_json, schema["allOf"])))
+        ]
         if any(s == FALSEY for s in schema["allOf"]):
             return FALSEY
         if all(s == TRUTHY for s in schema["allOf"]):
@@ -590,6 +604,8 @@ def merged(schemas: List[Any]) -> Optional[Schema]:
         return FALSEY
     out = schemas[0]
     for s in schemas[1:]:
+        if s == TRUTHY:
+            continue
         # If we have a const or enum, this is fairly easy by filtering:
         if "const" in out:
             if make_validator(s).is_valid(out["const"]):
@@ -611,60 +627,62 @@ def merged(schemas: List[Any]) -> Optional[Schema]:
             out["type"] = [
                 t for t in get_type(out) if t in tt or t == "integer" and "number" in tt
             ]
-            if not get_type(out):
+            out_type = get_type(out)
+            if not out_type:
                 return FALSEY
             for t, kw in TYPE_SPECIFIC_KEYS:
                 numeric = ["number", "integer"]
-                if t in get_type(out) or t in numeric and t in get_type(out) + numeric:
+                if t in out_type or t in numeric and t in out_type + numeric:
                     continue
                 for k in kw.split():
                     s.pop(k, None)
                     out.pop(k, None)
-        # TODO: keeping track of which elements are affected by which schemata
-        # while merging properties, patternProperties, and additionalProperties
-        # is a nightmare, so I'm just not going to try for now.  e.g.:
-        #    {"patternProperties": {".": {"type": "null"}}}
-        #    {"type": "object", "additionalProperties": {"type": "boolean"}}
-        # The the former requries null values for all non-"" keys, while the
-        # latter requires all-bool values.  Merging them should output
-        #    {"enum": [{}, {"": None}]}
-        # but dealing with this in the general case is a nightmare.
 
-        def diff_in_out(k: str) -> bool:
-            sval = s.get(k, object())
-            return k in s and sval != out.get(k, sval)
-
-        def diff_keys(k: str) -> bool:
-            return set(out.get(k, [])) != set(s.get(k, []))
-
-        if diff_in_out("patternProperties"):
-            # Do I want to compute regex intersections with optional anchors? No.
-            return None
-
-        if "additionalProperties" in out and (
-            diff_keys("properties") or diff_keys("patternProperties")
+        if out.get("patternProperties") != s.get("patternProperties") and (
+            out.get("properties") != s.get("properties")
+            or out.get("additionalProperties") != s.get("additionalProperties")
         ):
-            # If the known names or regex patterns vary at all, we'd have to merge the
-            # additionalProperties schema into some and it's just not worth it.
             return None
-        if diff_in_out("additionalProperties"):
-            m = merged([out["additionalProperties"], s["additionalProperties"]])
+
+        # With patternProperties set aside for now, it's reasonably easy to merge
+        # objects: we take all property names known to either side, and merge them
+        # with either the corresponding property or additionalProperties schema
+        # (and discard the properties dict if there weren't any).  Then, we merge
+        # the two additionalProperties (again, if there were any) and we're done!
+        #
+        # TODO: to  add patternProperties support, we 'just' check for matching
+        # pattern schemas in the loop and merge all of them (or a.P. none match);
+        # then add a second loop to merge each pP with the matching pP or aP
+        # (i.e. looking like the loop we have now), and then we're done again.
+        out_add = out.get("additionalProperties", {})
+        s_add = s.pop("additionalProperties", {})
+        if "properties" in out or "properties" in s:
+            out_props = out.setdefault("properties", {})
+            s_props = s.pop("properties", {})
+            for prop_name in set(out_props) | set(s_props):
+                m = merged(
+                    [out_props.get(prop_name, out_add), s_props.get(prop_name, s_add)]
+                )
+                if m is None:
+                    return None
+                out_props[prop_name] = m
+        if out_add or s_add:
+            m = merged([out_add, s_add])
             if m is None:
                 return None
             out["additionalProperties"] = m
 
-        if diff_in_out("properties"):
-            # TODO: this doesn't account for cases where patternProperties in one
-            # overlap with properties in the other.  It would be nice to try merging
-            # them in that case, or correctly bail out if we can't merge them.
-            op = out["properties"]
-            sp = s.pop("properties")
-            for k, v in sp.items():
-                if v != op.get(k, v):
-                    v = merged([op[k], v])
-                    if v is None:
-                        return None
-                op[k] = v
+        if "contains" in out and "contains" in s and out["contains"] != s["contains"]:
+            # OK, this is a tricky bit of logic.  If we are merging schemas with
+            # unequal "contains" keys, then we want to keep the *more* restrictive
+            # of the two as the contains key, so we can leverage that for generation.
+            contains = [out["contains"], s["contains"]]
+            contains.sort(key=upper_bound_instances)
+            out["contains"], _ = contains
+            out["allOf"] = out.get("allOf", []) + [{"contains": c} for c in contains]
+        if "allOf" in out and "allOf" in s:
+            # All our allOf schemas will be de-duplicated by canonicalise
+            out["allOf"] += s.pop("allOf")
         if "required" in out and "required" in s:
             out["required"] = sorted(set(out["required"] + s.pop("required")))
         for key in (
