@@ -256,6 +256,25 @@ def regex_patterns(draw: Any) -> str:
 
 REGEX_PATTERNS = regex_patterns()
 
+
+def json_pointers() -> st.SearchStrategy[str]:
+    """Return a strategy for strings in json-pointer format."""
+    return st.lists(
+        st.text(st.characters()).map(
+            lambda p: "/" + p.replace("~", "~0").replace("/", "~1")
+        )
+    ).map("".join)
+
+
+def relative_json_pointers() -> st.SearchStrategy[str]:
+    """Return a strategy for strings in relative-json-pointer format."""
+    return st.builds(
+        operator.add,
+        st.from_regex(r"0|[1-9][0-9]*", fullmatch=True),
+        st.just("#") | json_pointers(),
+    )
+
+
 # Via the `webcolors` package, to match the logic `jsonschema`
 # uses to check it's (non-standard?) "color" format.
 _WEBCOLOR_REGEX = "^#([a-fA-F0-9]{3}|[a-fA-F0-9]{6})$"
@@ -307,8 +326,8 @@ STRING_FORMATS = {
         name: prov.domains().map("https://{}".format)
         for name in ["uri", "uri-reference", "iri", "iri-reference", "uri-template"]
     },
-    "json-pointer": st.just(""),
-    "relative-json-pointer": st.just(""),
+    "json-pointer": json_pointers(),
+    "relative-json-pointer": relative_json_pointers(),
     "regex": REGEX_PATTERNS,
 }
 
@@ -356,6 +375,21 @@ def array_schema(schema: dict) -> st.SearchStrategy[List[JSONType]]:
         min_size = max(0, min_size - len(items))
         if max_size is not None:
             max_size -= len(items)
+
+        items_strats = [from_schema(s) for s in items]
+        additional_items_strat = from_schema(additional_items)
+
+        # If we have a contains schema to satisfy, we try generating from it when
+        # allowed to do so.  We'll skip the None (unmergable / no contains) cases
+        # below, and let Hypothesis ignore the FALSEY cases for us.
+        if "contains" in schema:
+            for i, mrgd in enumerate(merged([schema["contains"], s]) for s in items):
+                if mrgd is not None:
+                    items_strats[i] |= from_schema(mrgd)
+            contains_additional = merged([schema["contains"], additional_items])
+            if contains_additional is not None:
+                additional_items_strat |= from_schema(contains_additional)
+
         if unique:
 
             @st.composite  # type: ignore
@@ -366,11 +400,11 @@ def array_schema(schema: dict) -> st.SearchStrategy[List[JSONType]]:
                 def not_seen(elem: JSONType) -> bool:
                     return encode_canonical_json(elem) not in seen
 
-                for s in items:
-                    elems.append(draw(from_schema(s).filter(not_seen)))
+                for strat in items_strats:
+                    elems.append(draw(strat.filter(not_seen)))
                     seen.add(encode_canonical_json(elems[-1]))
                 extra_items = st.lists(
-                    from_schema(additional_items).filter(not_seen),
+                    additional_items_strat.filter(not_seen),
                     min_size=min_size,
                     max_size=max_size,
                     unique_by=encode_canonical_json,
@@ -380,14 +414,25 @@ def array_schema(schema: dict) -> st.SearchStrategy[List[JSONType]]:
 
             strat = compose_lists_with_filter()
         else:
-            fixed_items = st.tuples(*map(from_schema, items)).map(list)
-            extra_items = st.lists(
-                from_schema(additional_items), min_size=min_size, max_size=max_size
+            strat = st.builds(
+                operator.add,
+                st.tuples(*items_strats).map(list),
+                st.lists(additional_items_strat, min_size=min_size, max_size=max_size),
             )
-            strat = st.builds(operator.add, fixed_items, extra_items)
     else:
+        items_strat = from_schema(items)
+        if "contains" in schema:
+            contains_strat = from_schema(schema["contains"])
+            if merged([items, schema["contains"]]) != schema["contains"]:
+                # We only need this filter if we couldn't merge items in when
+                # canonicalising.  Note that for list-items, above, we just skip
+                # the mixed generation in this case (because they tend to be
+                # heterogeneous) and hope it works out anyway.
+                contains_strat = contains_strat.filter(make_validator(items).is_valid)
+            items_strat |= contains_strat
+
         strat = st.lists(
-            from_schema(items),
+            items_strat,
             min_size=min_size,
             max_size=max_size,
             unique_by=encode_canonical_json if unique else None,
